defmodule Eakins.UseCases.Reflection do
  @moduledoc """
  Abstracting operations across our image storage types
  """

  alias Eakins
  alias Eakins.UseCases.Reflection
  alias Eakins.Schema

  defmacro __using__(_) do
    quote do
      @behaviour unquote(__MODULE__)
      def fn_name(prefix, container_name, suffix \\ nil)

      def fn_name(prefix, container_name, suffix) do
        singular_container = Inflex.singularize(container_name)

        [prefix, singular_container, suffix]
        |> Enum.reject(&is_nil/1)
        |> Enum.join("_")
        |> String.to_existing_atom()
      end
    end
  end

  @type container_name :: String.t()

  @doc """
  Finds an image in an image container
  """
  @callback find_image(Schema.t(), container_name(), Images.key()) :: Images.image() | nil

  @doc """
  Creates a changeset that updates an image with the given key in the provided schema
  """
  @callback add_or_update_image(Schema.t(), String.t(), Eakins.key(), any) ::
              Ecto.Changeset.t()

  @doc """
  Creates a changeset that deletes the image with the given key
  """
  @callback delete_image(Schema.t(), String.t(), Ecto.Changeset.t(), Image.key()) ::
              Ecto.Changeset.t()

  @doc """
  Removes the given image from the underlying storage mechanism
  """
  @callback remove_image_from_storage(Schema.t(), String.t(), Images.image()) ::
              :ok | {:error, any}

  @doc """
  Returns true if the image container has the image with the given key
  """
  @callback has_image?(Schema.t(), String.t(), Image.key()) :: boolean

  @doc """
  Updates the image in the image container
  """
  @callback update_image(Schema.t(), String.t(), Ecto.Changeset.t(), Images.image()) ::
              Ecto.Changeset.t()

  @spec find_image(Schema.t(), String.t(), Images.key()) :: Images.image() | nil
  def find_image(%schema_module{} = schema, container_name, image_key) do
    module = implementation_module(schema_module, container_name)
    module.find_image(schema, container_name, image_key)
  end

  @spec add_or_update_image(Schema.t(), String.t(), Eakins.key(), any) ::
          Ecto.Changeset.t()
  def add_or_update_image(%schema_module{} = schema, container_name, image_key, uploaded_image) do
    module = implementation_module(schema_module, container_name)
    module.add_or_update_image(schema, container_name, image_key, uploaded_image)
  end

  @spec delete_image(Schema.t(), String.t(), Ecto.Changeset.t(), Image.key()) ::
          Ecto.Changeset.t()
  def delete_image(%schema_module{} = schema, container_name, changeset, image_key) do
    module = implementation_module(schema_module, container_name)
    module.delete_image(schema, container_name, changeset, image_key)
  end

  @spec remove_image_from_storage(Schema.t(), String.t(), Images.image()) :: :ok | {:error, any}
  def remove_image_from_storage(%schema_module{} = schema, container_name, old_image) do
    module = implementation_module(schema_module, container_name)
    module.remove_image_from_storage(schema, container_name, old_image)
  end

  @spec has_image?(Schema.t(), String.t(), Image.key()) :: boolean
  def has_image?(%schema_module{} = schema, container_name, image_key) do
    module = implementation_module(schema_module, container_name)
    module.has_image?(schema, container_name, image_key)
  end

  @spec update_image(Schema.t(), String.t(), Ecto.Changeset.t(), Images.image()) ::
          Ecto.Changeset.t()
  def update_image(%schema_module{} = schema, container_name, changeset, new_image) do
    module = implementation_module(schema_module, container_name)
    module.update_image(schema, container_name, changeset, new_image)
  end

  defp implementation_module(schema_module, container_name) do
    case schema_module.image_container_type(container_name) do
      :list -> Reflection.List
      :map -> Reflection.Map
    end
  end
end
