defmodule Eakins.UseCases.UpdateSchemaImage do
  @moduledoc """
  Updates one of the images in a schema's image map
  """
  alias Ecto.Changeset
  alias Ecto.Multi

  alias Eakins.{Images, Repo, Schema}
  alias Eakins.UseCases.Reflection

  import Eakins.Gettext

  require Logger

  use Ecto.Schema

  embedded_schema do
    field(:plug_upload, :map, virtual: true)
    field(:image_key, :string)
    field(:container_name, :string)
  end

  @type t :: %__MODULE__{}

  @spec new(Plug.Upload.t(), String.t(), Images.key()) :: t
  def new(plug_upload, container_name, image_key) do
    %__MODULE__{
      plug_upload: plug_upload,
      image_key: image_key,
      container_name: container_name
    }
  end

  @spec call(t(), Schema.t()) :: {:ok, Schema.t()} | {:error, Changeset.t()}
  def call(%__MODULE__{} = uploaded_image, schema) do
    container_name = uploaded_image.container_name
    image_key = uploaded_image.image_key
    plug_upload = uploaded_image.plug_upload

    old_image = Reflection.find_image(schema, container_name, image_key)

    Multi.new()
    |> Multi.update(
      :update_schema,
      Reflection.add_or_update_image(schema, container_name, image_key, plug_upload)
    )
    |> Multi.run(:remove_old_image, fn _repo, %{update_schema: schema} ->
      case Reflection.remove_image_from_storage(schema, container_name, old_image) do
        :ok ->
          {:ok, schema}

        {:error, reason} ->
          Logger.error("Could not remove old image #{inspect(reason)}")

          changeset =
            schema
            |> Changeset.change()
            |> Changeset.add_error(
              uploaded_image.image_key,
              gettext("Could not remove old image")
            )

          {:error, changeset}
      end
    end)
    |> Repo.transaction()
    |> case do
      {:ok, %{update_schema: schema}} ->
        {:ok, schema}

      {:error, _, %Changeset{} = changeset, _} ->
        {:error, changeset}
    end
  end
end
